from typing import Type
import torch
from anytraverse.utils import _typing as anyt
from anytraverse.utils.base.attention_mapping import PromptAttentionMapping
from anytraverse.utils.base.map_pooling import TraversabilityPooler, UncertaintyPooler
from anytraverse.utils.roi import RegionOfInterest
from anytraverse.utils.state import TraversalState, Threshold, AnyTraverseState
from anytraverse.utils.base.history import EncodingHistory
from anytraverse.utils.base.encoding import ImageEncoder
from anytraverse.utils.trav_pref import (
    update_traversability_preferences,
    parse_trav_pref_syntax,
    get_prompts,
    get_weights,
)


class AnyTraverse[TImage: anyt.Image]:
    """
    The offroad traversability pipeline discussed in "AnyTraverse: VLM-Based Offroad Traversability Framework with Human-in-the-Loop".
    """

    def __init__(
        self,
        prompt_attention_mapping: PromptAttentionMapping[TImage],
        traversability_pooler: Type[TraversabilityPooler],
        uncertainty_pooler: Type[UncertaintyPooler],
        init_traversability_preferences: anyt.TraversabilityPreferences,
        image_encoder: ImageEncoder[TImage],
        similarity_func: anyt.SimilarityFunction[anyt.Encoding, torch.Tensor],
        roi: RegionOfInterest,
        threshold: Threshold,
    ) -> None:
        """
        Initlializes the AnyTraverse pipeline for traversability segmentation.

        Args:
            prompt_attention_mapping (PromptAttentionMapping[TImage]):
                A `PromptAttentionMapping` module that takes in an image of type `TImage`
                and the prompts and outputs the `AttentionMap` for each prompt on the image.
            traversability_pooler (Type[TraversabilityPooler]):
                A pooler that takes in the attention maps and pools them into a traversability
                map using the provided traversability preferences.
            uncertainty_pooler (Type[UncertaintyPooler]):
                A pooler that takes in the attention maps and pools them into an uncertainty
                map using the provided traversability preferences.
            init_traversability_preferences (anyt.TraversabilityPreferences):
                The initial traversability preferences expressed as a `dict[str, float]`
                of the form `{"<prompt1>": <weight1>, ...}`
            image_encoder (ImageEncoder[TImage]):
                An image encoder that takes in and image of type `TImage` and outputs the
                encoded image as an `anyt.Encoding` (alias for `torch.Tensor`)
            similarity_func (SimilarityFunc[anyt.Encoding, torch.Tensor]):
                The simiilarity function to be used to match image encodings in the history.
                It should accept two `anyt.Encoding` encodings and output the similarity as
                a `torch.Tensor`.
                The type is an alias for `Callable[[anyt.Encoding, anyt.Encoding], torch.Tensor]`
            roi (RegionOfInterest):
                The `RegionOfInterest` module to extract the ROI in required maps.
            threshold (Threshold):
                The `dataclass` `Threshold`, containing the thresholds discussed in the paper
                - The reference scene similarity threshold (`ref_scene_sim`) for unseen scene human operator calls.
                - The ROI uncertainty threshold (`roi_uncertainty`) for unknown object human operator calls.
        """
        self._prompt_attention_mapping = prompt_attention_mapping
        self._traversability_pooler = traversability_pooler
        self._uncertainty_pooler = uncertainty_pooler
        self._traversability_preferences = init_traversability_preferences
        self._roi = roi
        self._similarity_func = similarity_func
        self._history = EncodingHistory(similarity_func=similarity_func)
        self._image_encoder = image_encoder
        self._traversal_state = TraversalState.OK
        self._threshold = threshold
        self._ref_scene_encoding: anyt.Encoding = torch.empty(0)
        self._current_scene_encoding: anyt.Encoding = torch.empty(0)

    @property
    def traversability_preferences(self) -> anyt.TraversabilityPreferences:
        """
        The traversability preferences as a `dict[str, float]` in
        the form `{"<prompt1>": <weight1>, ...}`
        """
        return self._traversability_preferences

    @traversability_preferences.setter
    def traversability_preferences(
        self, updated_traversability_preferences: anyt.TraversabilityPreferences
    ) -> anyt.TraversabilityPreferences:
        prompts, weights = (
            get_prompts(updated_traversability_preferences),
            get_weights(updated_traversability_preferences),
        )
        assert all(len(prompt) > 0 for prompt in prompts), (
            "Empty prompts not allowed in traversability preferences"
        )
        assert all(weight >= -1 and weight <= 1 for weight in weights), (
            "All weights in traversability preferences must be in range [-1, 1]"
        )
        self._traversability_preferences = updated_traversability_preferences
        return self._traversability_preferences

    def _create_maps(
        self, image: TImage
    ) -> tuple[
        list[anyt.PromptAttentionMap], anyt.TraversabilityMap, anyt.UncertaintyMap
    ]:
        """
        Creates the different maps to be used in the pipeline.

        Args:
            image (TImage): The image to generate the maps from.

        Returns:
            - attention_maps (list[PromptAttentionMap]):
                The attention maps on the image for each of the prompts, as a list.
            - traversability_map (TraversabilityMap):
                The traversability map showing the traversability (0-1) of each pixel in the image.
            - uncertainty_map (UncertaintyMap):
                The uncertainty map showing the uncertanty (0-1) of each pixel in the image.

            Each of the maps are `torch.Tensor` with shape `(H, W)`, the same as height and width of the image.
        """
        attention_maps = self._prompt_attention_mapping(
            x=image,
            prompts=get_prompts(
                traversability_preferences=self._traversability_preferences
            ),
        )
        traversability_map = self._traversability_pooler.pool(
            maps=attention_maps,
            traversability_preferences=self._traversability_preferences,
        )
        uncertainty_map = self._uncertainty_pooler.pool(
            maps=attention_maps,
            traversability_preferences=self._traversability_preferences,
        )
        return attention_maps, traversability_map, uncertainty_map

    def _update_traversability_preferences(
        self, delta_tau: anyt.TraversabilityPreferences
    ) -> anyt.TraversabilityPreferences:
        self._traversability_preferences = update_traversability_preferences(
            traversability_preferences=self._traversability_preferences,
            updates=delta_tau,
        )
        return self._traversability_preferences

    def _update_tau_and_history(
        self, delta_tau: anyt.TraversabilityPreferences = dict()
    ) -> anyt.HistoryElement[anyt.Encoding]:
        # Set reference scene encoding to current scene encoding
        self._ref_scene_encoding = self._current_scene_encoding
        # NOTE Technical - load to right device
        self._ref_scene_encoding = self._ref_scene_encoding.to(
            device=self._current_scene_encoding.device
        )
        # Update traversability preferences
        self._traversability_preferences = update_traversability_preferences(
            traversability_preferences=self._traversability_preferences,
            updates=delta_tau,
        )
        # Save new element in history
        new_history_element = self._history.add(
            key=self._current_scene_encoding,
            traversabilty_preferences=self._traversability_preferences,
        )
        return new_history_element

    def human_call(self, human_input: str) -> anyt.HistoryElement[anyt.Encoding]:
        """
        Perform a human operator call with human inputs for the traversability
        preferences expressed in the syntax:
        `<prompt1>: <weight1>[; <prompt2>: <weight2>; ...]`

        Args:
            human_input (str): The human input as a `str` in the syntax above.

        Returns:
            The updated traversability preferences in the AnyTraverse pipeline.
        """
        delta_tau = parse_trav_pref_syntax(syntax=human_input)
        return self._update_tau_and_history(delta_tau=delta_tau)

    def step(self, image: TImage) -> AnyTraverseState:
        """
        Perform one time step in the AnyTraverse pipeline, on the image of type `TImage`.
        The actual logic explained in the paper is implemented here.

        Args:
            image (TImage): The image to perform the functionality on.

        Returns:
            AnyTraverseState:
                The state of after passing the `image` through the AnyTraverse pipeline.
        """
        # Get the encoding of the current scene
        self._current_scene_encoding = self._image_encoder(x=image)

        # Set the reference scene encoding to the encoding of the first frame
        if self._ref_scene_encoding.numel() == 0:
            self._update_tau_and_history()

        # Create the different maps
        attention_maps, traversability_map, uncertainty_map = self._create_maps(
            image=image
        )

        # Extract the ROI from the traversability and uncertainty maps
        traversability_map_roi, roi_bbox_start, roi_bbox_end = self._roi.extract(
            mat=traversability_map
        )
        uncertainty_map_roi, _, _ = self._roi.extract(mat=uncertainty_map)

        # Check if human operator call required
        # Get the similarity with the reference scene
        ref_scene_similarity = self._similarity_func(
            self._current_scene_encoding, self._ref_scene_encoding
        )
        # Current scene not sufficiently similar to reference scene
        if ref_scene_similarity < self._threshold.ref_scene_similarity:
            # Search history for best match
            best_match_encoding, best_match_traversability_preferences = (
                self._history.find_best_match(query=self._current_scene_encoding)
            )
            # Is the best match similar enough?
            best_match_similarity = self._similarity_func(
                self._current_scene_encoding, best_match_encoding
            )
            if best_match_similarity >= self._threshold.ref_scene_similarity:
                # Successful hit in history
                # Update the reference scene embedding
                self._ref_scene_encoding = best_match_encoding
                # Update traversability preferences according to match in history
                self._update_traversability_preferences(
                    delta_tau=best_match_traversability_preferences
                )
                # Avoided redundant human operator call using history, good to go!
                self._traversal_state = TraversalState.OK
            else:
                # No good match found in history, unknown scene encountered
                self._traversal_state = TraversalState.UNKNOWN_SCENE
        elif uncertainty_map_roi.mean() > self._threshold.roi_uncertainty:
            # Too much uncertainty in the image ROI, unknown object detected
            self._traversal_state = TraversalState.UNKOWN_OBJ
        else:
            # No human operator call required, good to go!
            self._traversal_state = TraversalState.OK

        return AnyTraverseState(
            image_encoding=self._current_scene_encoding,
            attention_maps=attention_maps,
            traversability_map=traversability_map,
            uncertainty_map=uncertainty_map,
            traversability_map_roi=traversability_map_roi,
            uncertainty_map_roi=uncertainty_map_roi,
            ref_scene_similarity=ref_scene_similarity.item(),
            roi_bbox=(roi_bbox_start, roi_bbox_end),
            roi_traversability=traversability_map_roi.mean().item(),
            roi_uncertainty=uncertainty_map_roi.mean().item(),
            traversal_state=self._traversal_state,
            traversability_preferences=self.traversability_preferences,
        )



import torch
from PIL import Image as PILImage
from torchvision.transforms import Resize
from transformers import AutoProcessor, CLIPSegForImageSegmentation
from typing_extensions import override

from anytraverse.utils import _typing as anyt
from anytraverse.helpers.device import DEVICE
from anytraverse.utils.base.attention_mapping import PromptAttentionMapping


class CLIPSegAttentionMapping[TImage: anyt.Image](PromptAttentionMapping[TImage]):
    def __init__(
        self,
        device: torch.device = DEVICE,
        model_name: str = "mcmonkey/clipseg-rd64-refined-fp16",
    ) -> None:
        super().__init__()
        self._device = device
        self._processor = AutoProcessor.from_pretrained(
            model_name, use_fast=True, cache_dir="data/weights/clipseg"
        )
        self._model = CLIPSegForImageSegmentation.from_pretrained(
            model_name, cache_dir="data/weights/clipseg", device_map=str(device)
        )

    @override
    def __call__(
        self, x: TImage, prompts: str | list[str]
    ) -> list[anyt.PromptAttentionMap]:
        if not isinstance(x, PILImage.Image):
            x = PILImage.fromarray(x)
        if isinstance(prompts, str):
            prompts = [prompts]

        width, height = x.size
        resize = Resize(size=(height, width)).to(device=self._device)

        with torch.no_grad():
            inputs = self._processor(
                text=prompts,
                images=[x] * len(prompts),
                padding=True,
                return_tensors="pt",
            ).to(device=self._device)
            maps = [m for m in torch.sigmoid(resize(self._model(**inputs).logits))]
            return maps



from abc import ABC, abstractmethod

from anytraverse.utils import _typing as anyt


class BaseMapPooler[TInputMap, TPooledMap](ABC):
    """
    Base class for a map pooler.
    """

    @staticmethod
    @abstractmethod
    def pool(maps: list[TInputMap], *args, **kwargs) -> TPooledMap:
        pass


class PromptAttentionMapPooler[TOutputMap](
    BaseMapPooler[anyt.PromptAttentionMap, TOutputMap], ABC
):
    """
    Base pooler class for a map pooler that pools attention maps to a
    given type of map `TOutputMap`.
    """

    @staticmethod
    @abstractmethod
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> TOutputMap:
        pass


TraversabilityPooler = PromptAttentionMapPooler[anyt.TraversabilityMap]
UncertaintyPooler = PromptAttentionMapPooler[anyt.UncertaintyMap]



import torch
import numpy as np
from numpy import typing as npt
from anytraverse.utils import _typing as anyt
from anytraverse.utils.base.map_pooling import TraversabilityPooler
from anytraverse.utils.trav_pref import get_weights
from typing_extensions import override


class ProbabilisticTraversabilityPooler(TraversabilityPooler):
    @staticmethod
    @override
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> anyt.TraversabilityMap:
        """
        Probabilistic traversability pooler.

        Args:
            maps (list[PromptAttentionMap]):
                The prompt attention maps for each prompt, on the image.
                Each map is a `torch.Tensor` of shape `(H, W)`, where `H`
                is the height of the image and `W` is the width.
            traversability_preferences (TraversabilityPreferences):
                The traversability preferences of the AnyTraverse pipeline.

        Returns:
            TraversabilityMap:
                A `torch.Tensor` of shape `(H, W)`, showing traversability
                of each of the pixels in the image.
        """
        device = maps[0].device
        weights = get_weights(traversability_preferences=traversability_preferences)
        weights_arr = np.array(weights)
        weights_tensor = torch.tensor(weights_arr).reshape(-1, 1, 1).to(device=device)
        pos_inx, neg_inx = weights_arr > 0, weights_arr < 0
        z = weights_tensor * torch.stack(maps, dim=0)
        proba_trav = (1 - torch.prod(1 - z[pos_inx.flatten()], dim=0)) * torch.prod(  # type: ignore
            1 + z[neg_inx.flatten()],  # type: ignore
            dim=0,
            dtype=torch.float32,
        )

        return proba_trav


class WeightedMaxTraversabilityPooler(TraversabilityPooler):
    """
    The weighted max traversability pooler discussed in the paper.
    """

    @staticmethod
    @override
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> anyt.TraversabilityMap:
        # Create weighted maps
        weights = get_weights(traversability_preferences=traversability_preferences)
        weighted_maps = torch.stack(
            [weight * attn_map for attn_map, weight in zip(maps, weights)], dim=0
        )
        # Get the absolute weighted maps
        abs_weighted_maps = weighted_maps.abs()
        # For each pixel, which map has the highest absolute weighted value?
        _, max_abs_weighted_map_index = abs_weighted_maps.max(dim=0)
        # Create the pooled traversability map
        traversability_map: anyt.TraversabilityMap = (
            weighted_maps.gather(dim=0, index=max_abs_weighted_map_index.unsqueeze(0))
            .squeeze(0)
            .clip(min=0.0, max=1.0)
        )
        return traversability_map



from anytraverse.utils import _typing as typing
from anytraverse.utils.pipelines.paper import build_pipeline_from_paper


__all__ = ["typing", "build_pipeline_from_paper"]


def main():
    print("Welcome to AnyTraverse!")



from abc import ABC, abstractmethod
from anytraverse.utils import _typing as anyt


class BaseAttentionMapping[TInputMap, TPrompt, TOutputMap](ABC):
    """
    The base attention mapping module.

    Produces attention maps of type `TMapOutput` on inputs of type `TMapInput` using
    prompt(s) of type `TPrompt`.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def __call__(
        self, x: TInputMap, prompts: TPrompt | list[TPrompt]
    ) -> list[TOutputMap]:
        pass


class PromptAttentionMapping[TImage: anyt.Image](
    BaseAttentionMapping[TImage, anyt.Prompt, anyt.PromptAttentionMap]
):
    pass



import torch
import numpy as np
from numpy import typing as npt
from anytraverse.utils import _typing as anyt
from anytraverse.utils.base.map_pooling import UncertaintyPooler
from anytraverse.utils.trav_pref import get_weights
from typing_extensions import override


class ProbabilisticUncertaintyPooler(UncertaintyPooler):
    @staticmethod
    @override
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> anyt.UncertaintyMap:
        return (1 - torch.stack(maps, dim=0)).prod(dim=0).squeeze(0)


class NormalizedEntropyUncertaintyPooler(UncertaintyPooler):
    @staticmethod
    @override
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> anyt.UncertaintyMap:
        masks = torch.stack(maps, dim=0)
        weights: torch.Tensor = masks.max(dim=0).values
        probas: torch.Tensor = masks.softmax(dim=0)
        entropy: torch.Tensor = torch.sum(-probas * torch.log(probas), dim=0).squeeze(0)
        norm_entropy: torch.Tensor = entropy / np.log(masks.shape[0])
        return weights * norm_entropy


class InverseMaxProbabilityUncertaintyPooler(UncertaintyPooler):
    @staticmethod
    @override
    def pool(
        maps: list[anyt.PromptAttentionMap],
        traversability_preferences: anyt.TraversabilityPreferences,
    ) -> anyt.UncertaintyMap:
        return 1 - torch.stack(maps).max(dim=0).values



from abc import ABC, abstractmethod
from typing import Any, Sequence
from anytraverse.utils import _typing as anyt
from PIL import Image as PILImage
import torch
from anytraverse.helpers.device import DEVICE


class BaseEncoder[TInput](ABC):
    """
    The base encoder module.

    Encodes any input to a numeric representation, as `torch.Tensor`.
    """

    DIM: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    @abstractmethod
    def __call__(
        self, x: TInput | Sequence[TInput], *args: Any, **kwargs: Any
    ) -> anyt.Encoding:
        pass


class ImageEncoder[TImage: anyt.Image](BaseEncoder[TImage], ABC):
    """
    Base class for an image encoder module.
    """

    def __init__(self) -> None:
        super().__init__()


class HuggingFaceImageEncoder[TImage: anyt.Image](ImageEncoder[TImage], ABC):
    """
    Abstract base class for a wrapper for an image encoder from HuggingFace
    """

    def __init__(
        self,
        model_name: str,
        cache_dir: str | None = None,
        use_fast: bool = True,
        device: torch.device = DEVICE,
    ) -> None:
        super().__init__()
        self._model_name = model_name
        self._cache_dir = cache_dir
        self._use_fast_processor = use_fast
        self._device = device



import torch
from anytraverse.utils.pipelines.base import AnyTraverse
from anytraverse.utils.models.clipseg import CLIPSegAttentionMapping
from anytraverse.utils.models.clip import CLIPImageEncoder
from anytraverse.utils.modules.trav_pooling import WeightedMaxTraversabilityPooler
from anytraverse.utils.modules.unc_pooling import (
    InverseMaxProbabilityUncertaintyPooler,
)
from anytraverse.utils.roi import RegionOfInterest
from anytraverse.utils.state import Threshold
from anytraverse.utils import _typing as anyt
from PIL import Image as PILImage


def build_pipeline_from_paper(
    init_traversabilty_preferences: anyt.TraversabilityPreferences,
    ref_scene_similarity_threshold: float,
    roi_uncertainty_threshold: float,
    roi_x_bounds: tuple[float, float],
    roi_y_bounds: tuple[float, float],
) -> AnyTraverse[PILImage.Image]:
    """
    Builds the original pipeline from the paper.
    """
    return AnyTraverse[PILImage.Image](
        prompt_attention_mapping=CLIPSegAttentionMapping[PILImage.Image](),
        image_encoder=CLIPImageEncoder[PILImage.Image](
            model_name="openai/clip-vit-base-patch32"
        ),
        traversability_pooler=WeightedMaxTraversabilityPooler,
        uncertainty_pooler=InverseMaxProbabilityUncertaintyPooler,
        roi=RegionOfInterest(x_bounds=roi_x_bounds, y_bounds=roi_y_bounds),
        threshold=Threshold(
            ref_scene_similarity=ref_scene_similarity_threshold,
            roi_uncertainty=roi_uncertainty_threshold,
        ),
        init_traversability_preferences=init_traversabilty_preferences,
        similarity_func=torch.cosine_similarity,
    )



import torch
from transformers import AutoProcessor, CLIPVisionModelWithProjection
from anytraverse.utils.base.encoding import HuggingFaceImageEncoder
from anytraverse.utils import _typing as anyt
from typing import Literal, Sequence
from typing_extensions import override


class CLIPImageEncoder[TImage: anyt.Image](HuggingFaceImageEncoder[TImage]):
    """
    CLIP image encoder wrapper for use with AnyTraverse.
    """

    DIM = 512

    def __init__(
        self,
        model_name: Literal[
            "openai/clip-vit-base-patch32", "openai/clip-vit-base-patch16"
        ],
        cache_dir: str | None = None,
        use_fast: bool = True,
    ) -> None:
        super().__init__(model_name=model_name, cache_dir=cache_dir, use_fast=use_fast)
        self._model = CLIPVisionModelWithProjection.from_pretrained(
            self._model_name, cache_dir=self._cache_dir
        )
        self._processor = AutoProcessor.from_pretrained(
            self._model_name,
            use_fast=self._use_fast_processor,
            cache_dir=self._cache_dir,
        )

    @override
    def __call__(self, x: TImage | Sequence[TImage]) -> torch.Tensor:
        inputs = self._processor(images=x, return_tensors="pt")
        outputs = self._model(**inputs)
        return outputs.image_embeds



import torch
from torch.nn import functional as F
from transformers import AutoModel, AutoImageProcessor
from anytraverse.utils.base.encoding import HuggingFaceImageEncoder
from anytraverse.utils import _typing as anyt
from anytraverse.helpers.device import DEVICE


class NomicImageEncoder[TImage: anyt.Image](HuggingFaceImageEncoder[TImage]):
    """
    Nomic image encoding model
    """

    DIM = 768

    def __init__(self, device: torch.device = DEVICE) -> None:
        super().__init__(
            model_name="nomic-ai/nomic-embed-vision-v1.5",
            device=device,
            cache_dir="data/weights/nomic",
        )
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_name,
            use_fast=self._use_fast_processor,
            cache_dir=self._cache_dir,
        )
        self._model = AutoModel.from_pretrained(
            self._model_name,
            trust_remote_code=True,
            cache_dir=self._cache_dir,
            device_map=self._device,
        )

    def __call__(self, x: TImage | list[TImage]) -> anyt.Encoding:
        inputs = self._processor(x, return_tensors="pt").to(device=self._device)
        img_emb = self._model(**inputs).last_hidden_state
        return F.normalize(img_emb[:, 0], p=2, dim=1)



import torch
from numpy import typing as npt
from typing import TypeVar


TMat = TypeVar("TMat", torch.Tensor, npt.NDArray)


class RegionOfInterest:
    """
    Class to handle operations on the region of interest (ROI)
    of a given matrix
    """

    def __init__(
        self, x_bounds: tuple[float, float], y_bounds: tuple[float, float]
    ) -> None:
        self._x_bounds = x_bounds
        self._y_bounds = y_bounds

    def extract(self, mat: TMat) -> tuple[TMat, tuple[int, int], tuple[int, int]]:
        """
        Extracts the region of interest and the pixel coordinates of the ROI as a box.

        Args:
            mat (torch.Tensor | npt.NDArray): The image of map to find the ROI of as a
        """
        height, width = mat.shape[-2:]
        x_start, x_end = [int(width * bound) for bound in self._x_bounds]
        y_start, y_end = [int(height * bound) for bound in self._y_bounds]
        return (
            mat[y_start : y_end + 1, x_start : x_end + 1],
            (x_start, y_start),
            (x_end, y_end),
        )
